!***********************************************************************
! Module: ommosaic_mod
!
! Authors:  Christopher Brindle - SSC (MSSL)
!          
! Description: Mosaics OM images
! 
! Comments: Based on emosiac_mod.f90 written by R. West - SSC (Leicester)
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!
! Change History
!          July 9 2001 - CB Added check to see if any images completely 
!          overlap another one- if so the program outputs a warning message.
!          November 13 2001 CB - Ammended code to mosiac both low and high-
!          resolution images properly. Also modified to work if input files
!          don't have an exposure map. Modified to produce an exposure map
!          if required.
!          February 27,2008, VNY Added code for using the astrometry-correction
!          keywords (CRVAL1 and CRVAL1) together with the keyword POSCOROK
!          which was introduced to omatt (SSC-SPR-3707)
!          March, 11, 2008 VNY, The RA shifts of the individual images
!          that form the mosaic are multiplied by cos(DEC), SSC-SPR-3710  
!          Apr 17, 2008, VNY: interpolating of the RA and DEC offsets
!          (astrometry corrections) from the astrometry corrected exposures
!          to those having no astrometry corrections (e.g. when the correction
!          failed for some reason). In case the interpolation is impossible,
!          the offsets are taken from the nearest corrected exposure
!************************************************************************ 
module ommosaic_mod

  use dal
  use param  
  use errorHandling
  use constants
  use omnumerics_mod
  use sla
  use omutility_mod
!  use ominterpolate_mod
!  use splinefit_mod
!  use leastsquares_mod

  implicit none 
  
    private   

    !*********************************************
    ! A structure for storing the image parameters
    !*********************************************
    type ImageT
        character (256) :: filename
        real (single), dimension(:, :), pointer :: image
        real (single), dimension(:, :), pointer :: weight 
        real (double) :: xrpix, yrpix, xrval, yrval
        real (double) :: xdelt, ydelt, roll
        real(single) :: exposure
        integer :: datatype, dims(0 : 1), xmin, xmax, ymin, ymax
        integer(int32) :: binAx1
        real (double) :: sky2img(3, 3)
        logical :: poscorok
        logical :: posintrp
        real(double) :: xdeltMosaic, ydeltMosaic
        real(single) :: ra_off, dec_off
        real(double) :: ra_pnt, dec_pnt
        character(40) startObs
        character(40) endObs
    end type ImageT

    !***************************************
    ! The following structure will store 
    ! the observation details.
    !**************************************
    type OBSERVATIONDETAILS 
        character(40) telescope 
        character(40) instrument      
        character(40) observer      
        character(40) object
        character(40) filter
        character(40) obsID
        character(40) expID
        character(40) startObs
        character(40) endObs
        real(single) exposure
        logical :: poscorok
    end type OBSERVATIONDETAILS
    
    public ommosaic

   
contains 
!********************************************
! Subroutine ommosaic
!  
!********************************************
subroutine ommosaic

    type (ImageT), allocatable, dimension(:) :: input
    type (ImageT) :: output
   
    integer i, noFiles, noFilesPosCor, iPosCor
    character(256), allocatable, dimension(:) :: fileName
    character (256) :: outputFile
    integer :: maxIndex
    integer(int32), pointer, dimension(:) :: index
    real (single), dimension(:, :), pointer :: exposureImage
  
    real(single) :: maxPixelSize   
    integer(int32) :: exposureTime    
    integer(int32) :: noHighResImages
    logical :: exposureMap
    type (OBSERVATIONDETAILS) :: obsDetails 
    real(double) :: xrvalAverage, yrvalAverage, rollAverage, nAver
    real(single) :: raOffaver, decOffAver
    real(double) :: startMjd, endMjd
    real(double), allocatable, dimension(:) :: mjd, raOff,decOff
!    real(single), allocatable, dimension(:) :: mjdS, raOffS,decOffS
    integer, allocatable, dimension(:) :: interpolated

!    type(polintStruct) :: raPolint, decPolint
!    type(OneDSpline) :: raSpline
    real(single) :: polintError, polintX, polintRA, polintDEC

!    real(single), allocatable, dimension(:) :: weight
    integer(int8) :: maxIter
    real(single) :: gradient, gradientError, constant, constantError
    real(single) :: gradientD, gradientErrorD, constantD, constantErrorD
!    real(single) :: raLS, decLS
    integer :: j, lowFlag, highFlag, lowIndex, highIndex
    real(single) :: lowValue, highValue
    real(single) :: deltaMjd, deltaRA, deltaDEC, intRA, intDEC, intMjd
    logical :: intFlag
    

    noFiles = parameterCount("imagesets")
!    write(*,*) "L90 noFiles=", noFIles


    allocate(fileName(1 : noFiles))
    allocate(mjd(1 : noFiles)) 
    allocate(raOff(1 : noFiles)) 
    allocate(decOff(1 : noFiles))
    allocate(interpolated(1:noFIles))
    
    interpolated=-1 
    ! 0 corresponds to good (aspect corrected) data
    ! 1 corresponds to interpolated (not aspect corrected) data
    ! -1 corresponds to no data

    raOff=9.9D10
    decOff=9.9D10

    outputFile = stringParameter("mosaicedset")
 
    exposureMap = booleanParameter("exposuremap")   

    exposureTime = realParameter("exposure")
    !****************************************************       
    ! Open input files, read data arrays and WCS keywords
    !****************************************************
    
    allocate(input(1 : noFiles))
    allocate(index(1 : noFiles))
    maxPixelSize = 0.0
    maxIndex = 1
    noHighResImages = 0
     
    do i = 1, noFiles
        fileName(i) = trim(stringParameter("imagesets", i - 1))

        call readImage(input(i), fileName(i), obsDetails)
        startMjd=ModifiedJulianDate(input(i)%startObs)
        endMjd=ModifiedJulianDate(input(i)%endObs)
        mjd(i)=(startMjd+endMjd)/2.0_double
        raOff(i)=input(i)%ra_off
        decOff(i)=input(i)%dec_off

        !write(*,*) "L172 i=", i, " xrval=", input(i)%xrval, " yrval=", &
        !     input(i)%yrval, " ra_off=", input(i)%ra_off, " dec_off=", &
        !     input(i)%dec_off
!        write(*,*) "L138 i=", i, " start=", startMjd, " end=", endMjd, &
!             " mjd=", mjd(i)
!        write(*,*)  " raOff=", raOff(i), " decOff=", decOff(i)
!        write(*,*) "-----------------------"
       ! write(*,*) "L125 i=", i, " input ra_off, dec_off=", input(i)%ra_off, &
        !     input(i)%dec_off
        !write(*,*) "      start=", trim(input(i)%startObs), " end=", trim(input(i)%endObs)
        !************************************************
        ! Check to see if image is a high-resolution one
        !************************************************
        if (input(i)%binAx1 .eq. 0) then
            noHighResImages = noHighResImages + 1
            index(noHighResImages) = i
        end if
    end do

    noFilesPosCor=0
    do i = 1, noFiles

       if (input(i)%poscorok  .and. .not. input(i)%posintrp) then
          noFilesPosCor=noFilesPosCor+1
          interpolated(i)=0 ! no need interpolation
       endif
       !write(*,*) "L199 i=", i, " interpolated=", interpolated(i), &
       !     " poscorok=", input(i)%poscorok
    enddo
    
!    write(*,*) "L170 noFilesPosCor=", noFilesPosCor
    !if (noFilesPosCor .gt. 0) then 
       !allocate(mjdS(1 : noFilesPosCor)) 
       !allocate(raOffS(1 : noFilesPosCor)) 
       !allocate(decOffS(1 : noFilesPosCor))
       !allocate(weight(1:noFilesPosCor))
       !mjdS=0.0
       !raOffS=0.0
       !decOffS=0.0
       !weight=1.0
    !endif

    if (noHighResImages .gt. 0 .and. noHighResImages .lt. noFiles) then
        !*********************************************
        ! Convert these images to low-resolution ones
        !********************************************* 
        do i = 1, noHighResImages
            call ConvertToLowResImage(input(index(i)))
        end do 
    end if
    deallocate(index)


    mjd=mjd-51696.0D0

    !-----------------------------------------------
    ! Interpolating between two neighbouring values
    !----------------------------------------------
    !write(*,*) "L262 ************* Interpolating *******************"
    output%posintrp=.false.
    do i = 1, noFiles
       !write(*,*) "L233 i=", i, " interpolated=", interpolated(i), &
       !     " output posintrp=", output%posintrp
       if (interpolated(i) .lt. 0) then
          ! search for two neighbouring values
          lowValue=0.0
          highValue=0.0
          lowIndex=-1
          highIndex=-1
          intFlag=.false.
          ! look for the nearest low value
          do j=i,1,-1
             if (interpolated(j) .eq. 0) then
                lowIndex=j
                lowValue=raOff(j)
                exit
             endif
          enddo
          ! look for the nearest high value
          do j=i,noFiles
             if (interpolated(j) .eq. 0) then
                highIndex=j
                highValue=raOff(j)
                exit
             endif
          enddo
          
          !write(*,*) "     L261 lowIndex=", lowIndex, " highIndex=", highIndex
          intMjd=0.0
          intRA=0.0
          intDEC=0.0
          if (lowIndex .gt. 0 .and. highIndex .gt. 0) then ! interpolate             
             deltaMjd=mjd(highIndex)-mjd(lowIndex)
             deltaRA=raOff(highIndex)-raOff(lowIndex)
             deltaDEC=decOff(highIndex)-decOff(lowIndex)
             intFlag=.true.
             if (deltaMjd .eq. 0.0) then
                intRA=raOff(lowIndex)
                intDEC=decOff(lowIndex)
             else ! interpolate
                intMjd=mjd(i)-mjd(lowIndex)
                intRA=raOff(lowIndex)+intMjd*deltaRA/deltaMjd
                intDEC=decOff(lowIndex)+intMjd*deltaDEC/deltaMjd
             endif
          else ! use the nearest correct value
             if (lowIndex .gt. 0) then
                intRA=raOff(lowIndex)
                intDEC=decOff(lowIndex)
                intFlag=.true.
                !input(i)%posintrp=.true. ! image interpolated
             else
                if (highIndex .gt. 0) then
                   intRA=raOff(highIndex)
                   intDEC=decOff(highIndex)
                   intFlag=.true.
                endif
             endif
          endif
          !write(*,*) "L      292 intFlag=", intFlag
          !write(*,*) "L283 i=", i, " l/h Index=", lowIndex, highIndex, &
          !     " intMjd=", intMjd, " intRA=", intRA, " intDEC=", intDEC, &
          !     " intFlag=", intFlag
          if (intFlag) then
             !write(*,*) " setting interpolation flag to TRUE"
             input(i)%posintrp=.true. ! image interpolated
             output%posintrp=.true.
             interpolated(i)=1
             raOff(i)=intRA
             decOff(i)=intDEC
          endif
       endif
    enddo
    !write(*,*) "L307 ************* END Interpolating *******************"
! Correcting those images whouse shifts have been determined by the above  interpolation
!    write(*,*) "L287 ************* Correcting *******************"
    do i = 1, noFiles
       !write(*,*) "L312 i=", i, " posintrp=", input(i)%posintrp

       if (interpolated(i) .eq. 1) then ! add shifts
          input(i)%xrval=input(i)%xrval+raOff(i)/3600.0
          input(i)%yrval=input(i)%yrval+decOff(i)/3600.0  
          input(i)%ra_off=raOff(i)
          input(i)%dec_off=decOff(i)
          write(errstr,'(a, a, a)') "Adjusting ", trim(fileName(i)), &
               " to the nearest astrom.-corr. exp."
          call message(AppMsg, Verbose)
          call modifyAstrometryAttributes(input(i), fileName(i), obsDetails)
          !write(*,*) "L320"
       endif
    enddo
    
    xrvalAverage=0.0_double
    yrvalAverage=0.0_double
    rollAverage=0.0_double
    nAver=0.0_double

    raOffAver=0.0
    decOffAver=0.0

    !noFilesPosCor=0
    do i = 1, noFiles
       !write(*,*) "L334 i=", i
        !************************************************************
        ! Need to normalise the images to the same exposure time so 
        ! that they can be correctly coadded
        !************************************************************ 
        where(input(i)%image .gt. 0.0) input(i)%image = input(i)%image / input(i)%exposure
        !write(*,*) "L340"
        !**********************************************************
        ! Check the pixel size
        ! Check also whether the image was astrometrically corrected
        !***********************************************************

        if (input(i)%xdelt .gt. maxPixelSize) then
            maxIndex = i
            maxPixelSize = input(i)%xdelt  
        end if 
        ! add the corrected values to the average
        !write(*,*) "L311 i=", i, " poscoro=", input(i)%poscorok, &
        !     " raOff=",raOff(i), " decOff=", decOff(i), " mjd=", mjd(i)
        if (input(i)%poscorok .and. .not. input(i)%posintrp) then
           xrvalAverage=xrvalAverage+input(i)%xrval
           yrvalAverage=yrvalAverage+input(i)%yrval
           rollAverage=rollAverage+ input(i)%roll
           raOffAver=raOffAver+raOff(i)
           decOffAver=decOffAver+decOff(i)
           nAver=nAver+1.0_double
           !write(*,*) "L321 xrval=",input(i)%xrval, " yrval=",input(i)%yrval, &
           !     " nAver=", nAver, " raOffAver=", raOffAver, " decOffAver=", decOffAver
        endif
    end do 
    !write(*,*) "L363"
    iPosCor=0
    do i = 1, noFiles
       !write(*,*) "L248 i=", i, " poscorok=", input(i)%poscorok, " posintrp=", input(i)%posintrp
       if (input(i)%poscorok .and. .not. input(i)%posintrp) then
          iPosCor=iPosCor+1
          !mjdS(iPosCor)=mjd(i)
          !raOffS(iPosCor)=raOff(i)
          !decOffS(iPosCor)=decOff(i)          
       endif
    enddo
    !write(*,*) "L374 iPosCor=", iPosCor

    if (nAver .gt. 1.0_double) then
       xrvalAverage=xrvalAverage/nAver
       yrvalAverage=yrvalAverage/nAver
       rollAverage=rollAverage/nAver
       raOffAver=raOffAver/nAver
       decOffAver=decOffAver/nAver
    endif
    !write(*,*) "L383 xrvalAver=",  xrvalAverage, " yrvalAver=",yrvalAverage 

    ! ********************************************************
    ! If the number of aspect corrected exposures is greater
    ! than 1, then interpolate the raOff and decOff for those
    ! exposures that were not corrected
    !*********************************************************
    !if (iPosCor .gt. 1) then
       !call polintConstruct(noFilesPosCor, mjdS, raOffS, raPolint)
       !call polintConstruct(noFilesPosCor, mjdS, decOffS, decPolint)
!*************************************************************
! Least-square solution
!*************************************************************
!       maxIter=1
!       call LeastSquaresFit(noFilesPosCor, mjdS, raOffS, weight, maxIter, &
!            gradient, gradientError, constant, constantError)
!       call LeastSquaresFit(noFilesPosCor, mjdS, decOffS, weight, maxIter, &
!            gradientD, gradientErrorD, constantD, constantErrorD)
!       write(*,*) "L278 gradient=", gradient, " gradientError=", gradientError, &
!            " constant=", constant, " constantError=", constantError
!       write(*,*) "L280 gradientD=", gradientD, " gradientErrorD=", gradientErrorD, &
!            " constantD=", constantD, " constantErrorD=", constantErrorD
!       write(*,*) "----------------------"
!       do i=1, noFilesPosCor
!          raLS=constant+mjdS(i)*gradient
!          decLS=constantD+mjdS(i)*gradientD
!          write(*,*) "L286 i=", i, " mjdS=", mjdS(i), " raOffS=", raOffS(i), &
!               " raLS=", raLS, " decOffS=", decOffS(i), " decLS=", decLS
!       enddo
       !write(*,*) "=============================++"
       ! ******************************
       !do i = 1, noFiles
          !write(*,*) "...................."
          
          !polintX=mjd(i)
          !polintRA=polintInterpolate(polintX, raPolint, polintError) 
          !polintDEC=polintInterpolate(polintX, decPolint, polintError)
          !raLS=constant+mjd(i)*gradient
          !decLS=constantD+mjd(i)*gradientD
          !write(*,*) "L299 i=", i, " mjd=", mjd(i), " ra=   ", raOff(i), " dec =  ", decOff(i)
          !write(*,*) "L300 i=", i, " mjd=", mjd(i), " polRA=", polintRA, " polDec=", polintDEC
          !write(*,*) "L301 i=", i, " mjd=", mjd(i), "  raLS=", raLS,     "  decLS=", decLS
          !write(*,*) "..."

          !if (interpolated(i) .lt. 0) then ! replace the values
          !   interpolated(i)=1
          !   raOff(i)=raLS!polintRA
          !   decOff(i)=decLS!polintDEC
          !endif
       !enddo
       !call polintDestruct(decPolint)
       !call polintDestruct(raPolint)
    !else ! replace the missing values with the average 
    !   do i = 1, noFiles
    !      if (interpolated(i) .lt. 0) then ! replace the values
    !         interpolated(i)=1
    !         raOff(i)= raOffAver
    !         decOff(i)= decOffAver 
    !      endif
    !   enddo
    !endif

!    write(*,*) "-----------------------=============="
!    write(*,*) " L401 After interpolation:"
!    write(*,*) "-----------------------=============="
!    do i = 1, noFiles
!       write(*,*) "L272 i=", i, " mjd=", mjd(i), "  interpolated=", interpolated(i), &
!            " ra=", raOff(i), " dec=", decOff(i)
!    enddo
!    write(*,*) "------------------------============="

    !******************************************
    ! Check that the filters are all the same
    !******************************************
    call CheckFilters(noFiles, fileName, obsDetails)


    !*******************************************************
    ! Check that there are no overlapping windows
    ! If the images are from the same exposure 
    ! or from the same file then the subroutine CheckWindows
    ! issues a warning message (nothing else)
    !*******************************************************
    call CheckWindows(noFiles, fileName)     


    !**************************************************************
    ! 27.02.2008 VNY
    ! The outout image will have the the RA and DEC (xrval and yrval)
    ! corresponding to the  average xrval and yval from the input
    ! images (the averages of these values are is greater than zero)
    ! otherwise, use the values from the 
    ! image with the largest pixel size (maximal xdelt)
    !***************************************************************

    if (nAver .gt. 0.0_double) then
       !write(*,*) "L479"
       output%xrval = xrvalAverage
       output%xdelt = input(maxIndex)%xdelt
       output%yrval = yrvalAverage
       output%ydelt = input(maxIndex)%ydelt
       output%roll  = rollAverage
       output%poscorok=.true.
       obsDetails%poscorok=.true.
    else
       !write(*,*) "L488"
       output%xrval = input(maxIndex)%xrval
       output%xdelt = input(maxIndex)%xdelt
       output%yrval = input(maxIndex)%yrval
       output%ydelt = input(maxIndex)%ydelt
       output%roll  = input(maxIndex)%roll
       call warning("Non of the input images are astrometry-corrected.")
       output%poscorok=.false.
       obsDetails%poscorok=.false.
    endif

    !*********************************************************
    ! compute the differences in RA and DEC for each image
    ! with respect to the xrval and yrval values used for the 
    ! output image
    !**********************************************************
     do i = 1, noFiles
        input(i)%xdeltMosaic=input(i)%xrval-output%xrval
        input(i)%ydeltMosaic=input(i)%yrval-output%yrval
     enddo

    !*******************************************************
    ! Calculate size of output image (physical and in pixels)
    !*******************************************************
    call calculateOutputSize(noFiles, input, output)

    !************************************************                                                                          ! Loop through input images adding to the output
    !***********************************************

    allocate(exposureImage(0 : output%dims(0) - 1, 0 : output%dims(1) - 1))
    exposureImage = 0.0
  

    !write(*,*) "L220 bounds of exposureImage:"
    !write(*,*) "L221 1)=", lbound(exposureImage,1), " : ", ubound(exposureImage,1)
    !write(*,*) "L222 2)=", lbound(exposureImage,2), " : ", ubound(exposureImage,2)

    do i = 1, noFiles
        write(errstr,'(a, a, a)') "Adding image in file ", trim(fileName(i)), " to the mosaiced image"
      
        call message(AppMsg, Verbose)

        call addImage(input(i), output, exposureImage)
        
    end do
    deallocate(input)       
 
    !*****************************************************
    ! Normalise each pixel image by the sum of the 
    ! exposure times for each pixel
    !*****************************************************
    where (output%weight .gt. 0.0) output%image = output%image / real(output%weight)

    output%weight = exposureImage
    deallocate(exposureImage)

      
    where (output%image .gt. 0.0)  output%image = output%image * real(exposureTime)

    obsDetails%exposure = exposureTime
 
    call writeImage(output, outputFile, obsDetails, exposureMap)        


    !******************
    ! Release memory
    !******************
    deallocate(fileName)
    deallocate(mjd)
    deallocate(raOff)
    deallocate(decOff)
    deallocate(interpolated)

    !if (noFilesPosCor .gt. 0) then 
       !deallocate(mjdS)
       !deallocate(raOffS)
       !deallocate(decOffS)
       !deallocate(weight)
    !endif
end subroutine ommosaic

!*******************************************************************
! Calculate size of output image in physical units and in
! pixels. Achieves this by casting the coordinates of the corner
! pixels into the frames of the first image.
!*******************************************************************
subroutine calculateOutputSize(nInput, input, output)
    integer, intent(in) :: nInput
    type (ImageT), intent(inout) :: input(nInput), output
   

    ! Local variables
    integer :: i, j, k, xmin, xmax, ymin, ymax
    real (double) :: az, el, axvector(3), vlocal(3), vsky(3),&
         & xpix, ypix, outaz, outel

    ! Initialise bounds (in pixels)
    xmin = 0 
    xmax = 0 
    ymin = 0 
    ymax = 0

    !*****************************************************************
    ! Create local -> celestial transformation matrix for output image
    !*****************************************************************
    axvector(1) = output%xrval * DegToRad
    axvector(2) = output%yrval * DegToRad
    axvector(3)= output%roll * DegToRad
    call sla_deuler('ZYX', axvector(1), -axvector(2), axvector(3), &
         & output%sky2img)

    !****************************************************************
    ! Loop through the input images, casting each of the four corners
    ! into the frame of the output image. This allows us to find the
    ! bounding box of the input images in the output frame.
    !****************************************************************
    do i = 1, nInput
       axvector(1) = input(i)%xrval * DegToRad
       axvector(2) = input(i)%yrval * DegToRad
       axvector(3) = input(i)%roll  * DegToRad
       call sla_deuler('ZYX', axvector(1), -axvector(2), axvector(3),&
            & input(i)%sky2img)

       ! Initialise bounding box
       input(i)%xmin = 0
       input(i)%xmax = 0
       input(i)%ymin = 0
       input(i)%ymax = 0


       ! NB. There's a half-pixel shift in here somewhere which needs to
       ! be sorted out (thanks FITS people. No, really, thanks).
       !
             
       ! Loop through each corner
       do j = 0, 1
          az = (dble(input(i)%dims(0) * j) - input(i)%xrpix + 0.5_double) * input(i)%xdelt
          do k = 0, 1
             el = (dble(input(i)%dims(1) * k) - input(i)%yrpix + 0.5_double) * input(i)%ydelt

             !******************************************************* 
             ! Transform corner from local frame to output frame. I'm
             ! doing this the long-handed (and stupid) way as it's
             ! easier to get ones head around.
             !******************************************************

             ! Convert spherical -> cartesian
             call sla_dcs2c(az * DegToRad, el * DegToRad, vlocal)

             ! Transform from local to celestial (cartesian)
             call sla_dimxv(input(i)%sky2img, vlocal, vsky)
             
             ! Transform into output frame
             call sla_dmxv(output%sky2img, vsky, vlocal)

             ! Convert to back spherical coordinates
             call sla_dcc2s(vlocal, outaz, outel)

             ! Convert into pixels in the output frame
             xpix = outaz * RadToDeg / output%xdelt
             ypix = outel * RadToDeg / output%ydelt
     
             ! Check against global bounds
             xmin = min(xmin, floor(xpix))
             xmax = max(xmax, ceiling(xpix))
             ymin = min(ymin, floor(ypix))
             ymax = max(ymax, ceiling(ypix))

             ! Update local bounds
             input(i)%xmin = min(input(i)%xmin, floor(xpix))
             input(i)%xmax = max(input(i)%xmax, ceiling(xpix))
             input(i)%ymin = min(input(i)%ymin, floor(ypix))
             input(i)%ymax = max(input(i)%ymax, ceiling(ypix))
             
          end do
       end do
    end do

    ! Update reference pixel position of output image
    output%xrpix = -xmin
    output%yrpix = -ymin
    

    ! Calculate size of image, allocated array
    output%dims(0) = (xmax - xmin + 1)
    output%dims(1) = (ymax - ymin + 1)

    allocate(output%image(0 : output%dims(0) - 1, 0 : output%dims(1) - 1))
    output%image = -1.0   ! Was 0.0

    allocate(output%weight(0 : output%dims(0) - 1, 0 : output%dims(1) - 1))
    output%weight = 0.0

    ! Translate bounds of all input images
    do i = 1, nInput
        input(i)%xmin = input(i)%xmin - xmin
        input(i)%xmax = input(i)%xmax - xmin
        input(i)%ymin = input(i)%ymin - ymin
        input(i)%ymax = input(i)%ymax - ymin
    end do
     
end subroutine calculateOutputSize
!********************************************************
! Read image and associated WCS keywords from a dataset
!
!*********************************************************
subroutine readImage(image, imagefile, obsDetails)
    type (ImageT), intent(out) :: image
    character (*), intent(in) :: imagefile
    type (OBSERVATIONDETAILS), intent(inout) :: obsDetails 
    type (OBSERVATIONDETAILS) :: newObsDetails 
    
    ! Local variables
    integer, dimension(:), pointer :: dims
    type (DataSetT) :: set
    type (ArrayT) :: arr
    integer :: k
    character(40) :: startObs
    character(40) :: endObs
    logical, save :: first = .true. 
    logical :: poscorok, posintrp
    real(single) :: ra_off, dec_off

    ! Open dataset, find image array
    image%filename = imagefile

    set = dataset(imagefile, read)


    call GetObservationDetails(set, newObsDetails)
    startObs = newObsDetails%startObs
    endObs = newObsDetails%endObs

    if (first) then
        first = .false.
        obsDetails = newObsDetails
    else 
        k = CompareDates(obsDetails%startObs, startObs)
        
        if (k .lt. 0) obsDetails%startObs = startObs

        k = CompareDates(obsDetails%endObs, endObs)
        
        if (k .gt. 0) obsDetails%endObs = endObs
    end if

 
    arr = array(set, 'PRIMARY', read)
    
    ! Check dimensions of image
    if(numberOfDimensions(arr) .ne. 2) &
         call fatal('badDims', 'Only 2-D images are supported')

    ! Check the keyword POSCOROK
    poscorok=.false.
    if (hasAttribute(set, "POSCOROK")) then
       poscorok=booleanAttribute(set, "POSCOROK")
    else
       if (hasAttribute(arr, "POSCOROK")) then
          poscorok=booleanAttribute(arr, "POSCOROK")
       endif
    endif

    ! Check the keyword POSINTRP
    posintrp=.false.
    if (hasAttribute(set, "POSINTRP")) then
       posintrp=booleanAttribute(set, "POSINTRP")
       !write(*,*) "L760 posintrp=", posintrp
    else
       if (hasAttribute(arr, "POSINTRP")) then
          posintrp=booleanAttribute(arr, "POSINTRP")
          !write(*,*) "L764 posintr=", posintrp
       endif
    endif

    ! Check the keywords RA_OFF and DEC_OFF
    ra_off=9.9E10
    dec_off=9.9E10
    if (hasAttribute(set, "RA_OFF")) then
       ra_off=real32Attribute(set, "RA_OFF")
    else
       if (hasAttribute(arr, "RA_OFF")) then
          ra_off=real32Attribute(arr, "RA_OFF")
       endif
    endif
    if (hasAttribute(set, "DEC_OFF")) then
       dec_off=real32Attribute(set, "DEC_OFF")
    else
       if (hasAttribute(arr, "DEC_OFF")) then
          dec_off=real32Attribute(arr, "DEC_OFF")
       endif
    endif

    if (hasAttribute(set, "RA_PNT")) then
       image%ra_pnt=real64Attribute(set, "RA_PNT")
    else
       if (hasAttribute(arr, "RA_PNT")) then
          image%ra_pnt=real64Attribute(arr, "RA_PNT")
       endif
    endif
    if (hasAttribute(set, "DEC_PNT")) then
       image%dec_pnt=real64Attribute(set, "DEC_PNT")
    else
       if (hasAttribute(arr, "DEC_PNT")) then
          image%dec_pnt=real64Attribute(arr, "DEC_PNT")
       endif
    endif

    !write(*,*) "L745 poscorok=", poscorok
    image%poscorok=poscorok
    image%ra_off=ra_off
    image%dec_off=dec_off
    image%startObs=startObs
    image%endObs=endObs
    image%posintrp=posintrp ! the RA and DEC positions are not yet interpolated

    dims=>dimensions(arr)
    allocate(image%image(0 : dims(0) - 1, 0 : dims(1) - 1))
    allocate(image%weight(0 : dims(0) - 1, 0 : dims(1) - 1))
    
    image%dims = dims

    ! Read WCS keywords
    image%xrpix = real64Attribute(arr, 'CRPIX1')
    image%xrval = real64Attribute(arr, 'CRVAL1')
    image%xdelt = real64Attribute(arr, 'CDELT1')
    image%yrpix = real64Attribute(arr, 'CRPIX2')
    image%yrval = real64Attribute(arr, 'CRVAL2')
    image%ydelt = real64Attribute(arr, 'CDELT2')
   
    image%exposure = real32Attribute(set, "EXPOSURE")
    image%binAx1 = int32Attribute(set, "BINAX1")

  
    image%roll = 0.0

   ! Perform type conversion on image data
    image%datatype = datatype(arr)
    call readImageAsReal32(arr, image%image)



    !*******************************
    ! Read in the exposure image
    !*******************************
    if (hasBlock(set, "EXPOSURE")) then
        arr = array(set, 'EXPOSURE', read)
       
 
        call readImageAsReal32(arr, image%weight)
        
        where(image%weight .gt. 0.0)  ! NB was image%image 
            image%weight = 1.0
        elsewhere
            image%weight = 0.0
        end where   
         
    else
        where(image%image .ge. 0.0) 
            image%weight = 1.0
          
            
        elsewhere
            image%image = -1.0
            image%weight = 0.0
            
        end where
    end if
  
     

    ! Free the dataset
    call release(arr)
    call release(set)


end subroutine readImage

!************************************************************
! Modify the astrometry attributes (poscorok, ra_off, dec_off)
!
!*************************************************************
subroutine modifyAstrometryAttributes(image, imagefile, obsDetails)
    type (ImageT), intent(inout) :: image
    character (*), intent(in) :: imagefile
    type (OBSERVATIONDETAILS), intent(inout) :: obsDetails 
    type (OBSERVATIONDETAILS) :: newObsDetails 
    
    ! Local variables
    integer, dimension(:), pointer :: dims
    type (DataSetT) :: set
    type (ArrayT) :: arr
    integer :: k
    character(40) :: startObs
    character(40) :: endObs
    logical, save :: first = .true. 
    logical :: poscorok
    real(single) :: ra_off, dec_off
    real(double) :: ra_pnt, dec_pnt

    ra_off=0.0
    dec_off=0.0
    ra_pnt=99999.0_double
    dec_pnt=99999.0_double
    ! Open dataset, find image array
    !image%filename = imagefile
!    write(*,*) "L888 image filename=", trim(image%filename)
!    write(*,*) "L889 image ra_off=", image%ra_off, " dec_off=", image%dec_off
!    write(*,*) "L890 image ra_pnt=", image%ra_pnt, " dec_pnt=", image%dec_pnt
!    write(*,*) "L891 image xrval=", image%xrval, " yrval=", image%yrval

    set = dataset(imagefile, modify)

!    set = dataset(imagefile, read)

!    call GetObservationDetails(set, newObsDetails)
!    startObs = newObsDetails%startObs
!    endObs = newObsDetails%endObs
!
!    if (first) then
!        first = .false.
!        obsDetails = newObsDetails
!    else 
!        k = CompareDates(obsDetails%startObs, startObs)
!        
!        if (k .lt. 0) obsDetails%startObs = startObs!
!
!        k = CompareDates(obsDetails%endObs, endObs)
!        
!        if (k .gt. 0) obsDetails%endObs = endObs
!    end if
!
! 
!    write(*,*) "L880"
    arr = array(set, 'PRIMARY', modify)
!    arr = array(set, 'PRIMARY', read)
!    write(*,*) "L882"
!    
!    ! Check dimensions of image
!    if(numberOfDimensions(arr) .ne. 2) &
!         call fatal('badDims', 'Only 2-D images are supported')!
!
!    ! Check the keyword POSCOROK
!    poscorok=.false.
!    if (hasAttribute(set, "POSCOROK")) then
!       poscorok=booleanAttribute(set, "POSCOROK")
!    else
!       if (hasAttribute(arr, "POSCOROK")) then
!          poscorok=booleanAttribute(arr, "POSCOROK")
!       endif
!    endif
!!!
!
!    ! Check the keywords RA_OFF and DEC_OFF
!    ra_off=9.9E10
!    dec_off=9.9E10
    !***********************************
    ! First delete the old attributes
    !************************************
    if (hasAttribute(arr, "POSINTRP")) then
       call deleteAttribute(arr,"POSINTRP")
    endif

    if (hasAttribute(arr, "POSCOROK")) then
       call deleteAttribute(arr,"POSCOROK")
    endif

    if (hasAttribute(arr, "DEC_OFF")) then
       call deleteAttribute(arr,"DEC_OFF")
    endif

    if (hasAttribute(arr, "RA_OFF")) then
       call deleteAttribute(arr,"RA_OFF")
    endif

    if (hasAttribute(arr, "DEC_PNT")) then
       call deleteAttribute(arr,"DEC_PNT")
    endif

    if (hasAttribute(arr, "RA_PNT")) then
       call deleteAttribute(arr,"RA_PNT")
    endif

    if (hasAttribute(arr, "CRVAL2")) then
       call deleteAttribute(arr,"CRVAL2")
    endif

    if (hasAttribute(arr, "CRVAL1")) then
       call deleteAttribute(arr,"CRVAL1")
    endif

    !*************************************
    ! Set new values for these attributes
    !*************************************
    
    call setAttribute(arr,"CRVAL1", image%xrval, 'deg', "Coord. of pointing direction for X")     
    call setAttribute(arr,"CRVAL2", image%yrval, 'deg', "Coord. of pointing direction for Y") 

    call setAttribute(arr,"RA_PNT", image%ra_pnt, 'deg', "RA pointing direction")     
    call setAttribute(arr,"DEC_PNT", image%ra_pnt, 'deg', "DEC pointing direction ")     

    call setAttribute(arr,"RA_OFF", image%ra_off, 'arcsec', "RA astrom. corr. shift ")     
    call setAttribute(arr,"DEC_OFF", image%ra_off, 'arcsec', "DEC astrom. corr. shift ") 
    
    call setAttribute(arr,"POSCOROK", .true., "Image is astrom. corrected")     
    call setAttribute(arr,"POSINTRP", .true., "Astrom corr. by interpolation")     



!    write(*,*) "L931 ra_pnt=", ra_pnt, " dec_pnt=", dec_pnt

!    if (hasAttribute(set, "DEC_OFF")) then
!       dec_off=real32Attribute(set, "DEC_OFF")
!    else
!       if (hasAttribute(arr, "DEC_OFF")) then
!          dec_off=real32Attribute(arr, "DEC_OFF")
!       endif
!    endif
!    write(*,*) "L745 poscorok=", poscorok
!    image%poscorok=poscorok
!    image%ra_off=ra_off
!    image%dec_off=dec_off
!    image%startObs=startObs
!    image%endObs=endObs
!    image%posintrp=.false. ! the RA and DEC positions are not yet interpolated
!
!    dims=>dimensions(arr)
!    allocate(image%image(0 : dims(0) - 1, 0 : dims(1) - 1))
!    allocate(image%weight(0 : dims(0) - 1, 0 : dims(1) - 1))
!    
!    image%dims = dims
!
!    ! Read WCS keywords
!    image%xrpix = real64Attribute(arr, 'CRPIX1')
!    image%xrval = real64Attribute(arr, 'CRVAL1')
!    image%xdelt = real64Attribute(arr, 'CDELT1')
!    image%yrpix = real64Attribute(arr, 'CRPIX2')
!    image%yrval = real64Attribute(arr, 'CRVAL2')
!    image%ydelt = real64Attribute(arr, 'CDELT2')
!   
!    image%exposure = real32Attribute(set, "EXPOSURE")
!    image%binAx1 = int32Attribute(set, "BINAX1")
!
!  
!    image%roll = 0.0
!
!   ! Perform type conversion on image data
!    image%datatype = datatype(arr)
!    call readImageAsReal32(arr, image%image)
!!
!!
!
!   !*******************************
!    ! Read in the exposure image
!    !*******************************
!    if (hasBlock(set, "EXPOSURE")) then
!        arr = array(set, 'EXPOSURE', read)
!       
! 
!        call readImageAsReal32(arr, image%weight)
!        
!        where(image%weight .gt. 0.0)  ! NB was image%image 
!            image%weight = 1.0
!        elsewhere
!            image%weight = 0.0
!        end where   
!         
!    else
!        where(image%image .ge. 0.0) 
!            image%weight = 1.0
!          
!            
!        elsewhere
!            image%image = -1.0
!            image%weight = 0.0
!            
!        end where
!    end if
!  
!     
!
    ! Free the dataset
    call release(arr)
    call release(set)


end subroutine modifyAstrometryAttributes

!********************************************
! Add an image to the output image
!
!********************************************
subroutine addImage(input, output, exposureImage)
    type (ImageT), intent(in) :: input
    type (ImageT), intent(inout) :: output
    real(single), intent(inout), dimension(0:, 0:) :: exposureImage  
 
    ! Local variables
    integer :: i, j, xpix, ypix
    real (double) :: iaz, iel, oel, vlocal(3), vsky(3), cosdelta
    real(double), dimension(:), allocatable :: oaz  
    !****************************************************************
    ! Loop over every pixel in the output image, transform from local
    ! coordinates to sky coordinates, transform from sky coordinates
    ! local coordinates in the input image, pick up pixel value and
    ! copy to output pixel
    !
    ! The matrix stuff here could be optimised quite a bit I think
    !***************************************************************
    cosdelta=cos(input%yrval*DegToRad)
!    write(*,*) "L544 dim0=", output%dims(0), " dim1=", output%dims(1), &
!         " xdelt=", output%xdelt, " DegToRad=", DegToRad
    allocate(oaz(1 :  output%dims(0)))
    forall (i = 1 :  output%dims(0)) oaz(i) = &
               ((dble(i) - output%xrpix) * output%xdelt) * DegToRad
!    write(*,*) "L548 oaz(283)=", oaz(283), " xrpix=", output%xrpix, &
!         " yrpix=", output%yrpix

    do j = 0, output%dims(1) - 1
        oel = ((dble(j + 1) - output%yrpix) * output%ydelt) * DegToRad

        do i = 0, output%dims(0) - 1
  

            !***********************************
            ! Convert spherical -> cartesian
            !***********************************
            call sla_dcs2c(oaz(i + 1), oel, vlocal)
!            if (j .eq. 282 .and. i .eq. 835) then
!               write(*,*) "L561 oaz(i+1)=", oaz(i+1), " i=", i
!            endif
            !****************************************************
            ! Transform from local to celestial (cartesian)
            ! sla_dimxv applies a 3d reverse rotation 
            ! sky2img is the rotation matrix, vlocal the
            ! vector to be rotated and vsky the resultant vector   
            !****************************************************
            call sla_dimxv(output%sky2img, vlocal, vsky)

            !********************************
            ! Transform into output frame
            ! sla_dmxv applies a 3-d rotation
            ! sky2img is the rotation matrix,
            ! vsky the vector to be rotated and
            ! vlocal the resultant vector
            !********************************
            call sla_dmxv(input%sky2img, vsky, vlocal)

            !*******************************************
            ! Convert to back spherical coordinates
            ! sla_dcc2s converts cartesian to spherical
            !*******************************************
            call sla_dcc2s(vlocal, iaz, iel)

            !*********************************************************
            ! Convert into pixels in the output frame
            ! NB - xdelt is the increment per pixel in the x-direction
            !      ydelt is the increment per pixel in the y-direction
            !      xrpix is the x-pixel reference point
            !      yrpix is the y-pixel reference point
            !**********************************************************
            !xpix = nint(iaz * RadToDeg / input%xdelt + input%xrpix) - 1
            !ypix = nint(iel * RadToDeg / input%ydelt + input%yrpix) - 1
!***************************************************************************
! 27.02.2008 adding the RA and DEC shifts of the partial image with respect
! to the output combined image
!***************************************************************************
! 11.03.2008 the shifts along the RA axis must be multiplied by cos(DEL)
! Although the effect of this is very small, for near-pole images this will 
! reduce the differences in RA 
!***************************************************************************
            !xpix = nint((iaz * RadToDeg - input%xdeltMosaic)/ input%xdelt +  &
            !     input%xrpix) - 1
            xpix = nint(iaz * RadToDeg/input%xdelt - & 
                 input%xdeltMosaic * cosdelta / input%xdelt +  &
                 input%xrpix) - 1
            ypix = nint((iel * RadToDeg - input%ydeltMosaic)/ input%ydelt + &
                 input%yrpix) - 1

            !******************************************
            ! Check against bounds of input image
            !******************************************
            if(xpix .gt. 0 .and. xpix .lt. input%dims(0) .and. &
                   ypix .gt. 0 .and. ypix .lt. input%dims(1)) then

                if (output%image(i, j) .lt. 0.0 .and. &
                    input%image(xpix, ypix) .ge. 0.0) output%image(i, j) = 0.0  ! NB was -1
                if (input%image(xpix, ypix) .gt. 0.0) then

                    output%image(i, j) = output%image(i, j) + &
                     & input%image(xpix, ypix)
                end if

                !************************************************
                ! Sum up the weighting image for the normalisation
                !************************************************
                output%weight(i, j) = output%weight(i, j) + input%weight(xpix, ypix)

                !************************************************
                ! Sum up the exposure image for the output file
                !************************************************
                if (input%weight(xpix, ypix) .gt. 0.0) &
                    exposureImage(i, j) = exposureImage(i, j) + input%exposure  
                
            end if

        end do
    end do
    deallocate(oaz)
end subroutine addImage
!**************************************************
! Write image to a new file
!**************************************************
subroutine writeImage(image, filename, obsDetails, exposureMap)
    type (imageT), intent(inout) :: image
    character (len=*), intent(in) :: filename
    type (OBSERVATIONDETAILS), intent(in) :: obsDetails
    logical, intent(in) :: exposureMap

    ! Local variables
    type (DataSetT) :: set
   
   
    ! Create new dataset
    set = dataSet(filename, CREATE)


    !*******************************************
    ! Add the fits header information
    !*******************************************
    call AddFitsKeyWords(set, obsDetails)

    !********************************
    ! Add the image array
    !********************************
    
    where(image%weight .eq. 0.0) 
         image%weight = -1.0
         image%image = -1.0
    end where



    call AddImageArray(set, "PRIMARY", image, image%image, .true.)


    !***************************************
    ! Add the exposure array (if required)
    !***************************************
    if (exposureMap) then
        call AddImageArray(set, "EXPOSURE", image, image%weight, .false.)
    end if
 

    call release(set)
    
end subroutine writeImage

!********************************************************************
! Read an image as Real32. Assumes pointer points to correctly sized
! array of doubles.
!*******************************************************************
subroutine readImageAsReal32(arr, image)   ! Was real 64
    type (ArrayT), intent(in) :: arr
    real (single), pointer, dimension(:,:) :: image

    ! Local variables
    real (single), pointer :: f32array(:,:)
    real (double), pointer :: f64array(:,:)
    integer (int16), pointer :: i16array(:,:)
    integer (int32), pointer :: i32array(:,:)
   

    ! Perform type conversion on image data
   

    select case (datatype(arr))
        case (integer16)
            i16array=>int16Array2Data(arr)
            image = i16array
        case (integer32)
            i32array=>int32Array2Data(arr)
            image = i32array
        case (real32)
            f32array=>real32Array2Data(arr)
            image = f32array
        case (real64)
            f64array=>real64Array2Data(arr)
            image = f64array
        case default
            call error('badType', 'Image is of unsupported data type')
    end select
    
end subroutine readImageAsReal32

!**************************************************
! CheckFilters
! Checks that the filters in the files are all
! identical
!**************************************************
subroutine CheckFilters(nosw, oswFile, obsDetails)
    integer, intent(in) :: nosw
    integer, parameter :: nFilt = 7
    character(*), intent(in), dimension(1 : nosw) :: oswFile
    Type (OBSERVATIONDETAILS), intent(inout) :: obsDetails
    character(20), allocatable, dimension(:) :: filter
    character(20) :: oswFilter
    character(100) :: string
    integer :: numFilters 
    integer :: i, j
    type (DataSetT) :: oswDataSet
    logical :: found

    allocate(filter(1 : nosw))
 

    numFilters = 1
    do i = 1 , nosw
        oswDataSet = DataSet(oswFile(i),read)
        oswFilter = trim(stringAttribute(oswDataSet, "FILTER"))
        
        if (i .eq. 1) then
            filter(1) = oswFilter
        else
            found = .false. 
            do j = 1, numFilters
        

                if (oswFilter .eq. filter(j)) then
                    found = .true.
                    exit
                end if
            end do
            if (.not. found) then
                numFilters = numFilters + 1 
                filter(numFilters) = oswFilter
            end if 

      
        end if    
        call release(oswDataSet)
   
 
    end do
  
    if (numFilters .gt. 1) then
       string = "("
       do i = 1, numFilters - 1
           string = trim(string) // trim(filter(i)) // ", "
       end do
       
       string = trim(string) // trim(filter(numFilters)) // ")"

       call warning("You are combining images from different filters!", trim(string))
       obsDetails%filter = "VARIOUS"
    end if

    deallocate(filter)

end subroutine CheckFilters
!********************************************************************
! Subroutine CheckWindows
! Purpose - To check to see if any image overlaps with another one.
! If so, outputs a warning message.
!*******************************************************************
subroutine CheckWindows(nosw, oswFile)
    integer, intent(in) :: nosw
    character(*), intent(in), dimension(1 : nosw) :: oswFile
    character(20), allocatable, dimension(:) :: ExposureNo
    character(20) :: exposureID
    character(100) :: string
    integer :: i, j
    type (DataSetT) :: oswDataSet
    
    allocate(exposureNo(1 : nosw))

    do i = 1, nosw

        oswDataSet = DataSet(oswFile(i),read)

        !****************************
        ! Get the exposure No
        !****************************
        exposureID = trim(stringAttribute(oswDataSet, "EXP_ID"))
        read(exposureID(11 : 13), *) exposureNo(i)
    
        call release(oswDataSet)

        do j = 1, i - 1
          

            if (oswFile(i) .eq. oswFile(j)) then     
                string = "(" // trim(oswFile(i)) // ")" 
                        call warning("You are combining identical images!", trim(string))
            else
                if (exposureNo(i) .eq. exposureNo(j)) then
                    string = "(" // trim(oswFile(i)) // ", " // trim(oswFile(j)) // ")"
                    call warning("You are combining images from the same exposure!", trim(string))
                end if
            end if
           
        end do
    end do
    
    deallocate(exposureNo)
    
end subroutine CheckWindows
!********************************************
! ConvertToLowResImage
! Converts an image from high-resolution
! to low-resolution
!********************************************
subroutine ConvertToLowResImage(image)

    type (imageT), intent(inout) :: image
    type (imageT) :: lowResImage
    integer :: i, j, i1, j1, k, l
    integer :: count

    if ((image%dims(0) / 2) * 2 .eq. image%dims(0)) then 
        lowResImage%dims(0) = image%dims(0) / 2 
    else
        lowResImage%dims(0) = image%dims(0) / 2 + 1 
    end if  

    if ((image%dims(1) / 2) * 2 .eq. image%dims(1)) then 
        lowResImage%dims(1) = image%dims(1) / 2 
    else
        lowResImage%dims(1) = image%dims(1) / 2 + 1 
    end if  


    !*************************************
    ! Create the new low-res image array
    !*************************************
    allocate(lowResImage%image(0 : lowResImage%dims(0) - 1, 0 : lowResImage%dims(1) - 1))
    lowResImage%image = 0.0
    allocate(lowResImage%weight(0 : lowResImage%dims(0) - 1, 0 : lowResImage%dims(1) - 1))
    lowResImage%weight = 0.0

    !******************************************************
    ! Need to co-add the four pixels of the high-res image
    ! to produce a low-res pixel
    !******************************************************

    do i = 0, image%dims(0) - 1, 2
        i1 = i / 2        
        do j = 0, image%dims(1) - 1, 2
            j1 = j / 2
            count = 0            
            do k = i, min(i + 1, image%dims(0) - 1) 
                do l = j, min(j + 1, image%dims(1) - 1)
                    lowResImage%image(i1, j1) = lowResImage%image(i1, j1) + image%image(k, l)
                    lowResImage%weight(i1, j1) = lowResImage%weight(i1, j1) + image%weight(k, l)
                    count = count + 1 
                end do
          
            end do 
            !********************************************************************
            ! Normalize this pixel by no of pixels (usually 4) used in averaging
            !********************************************************************
            lowResImage%weight(i1, j1) = lowResImage%weight(i1, j1) / real(count)

        end do
    end do
    
    image%dims(0) = lowResImage%dims(0) 
    image%dims(1) = lowResImage%dims(1)

    image%xdelt = image%xdelt * 2.0_double
   
    image%ydelt = image%ydelt * 2.0_double
     
    image%xrpix = image%xrpix * 0.5_double
    image%yrpix = image%yrpix * 0.5_double
    
    deallocate(image%image)
    deallocate(image%weight)

    allocate(image%image(image%dims(0), image%dims(1)))
    allocate(image%weight(image%dims(0), image%dims(1)))

    image%image = lowResImage%image
    image%weight = lowResImage%weight

    deallocate(lowResImage%image)
    deallocate(lowResImage%weight)

end subroutine ConvertToLowResImage
!****************************************
! Subroutine AddImageArray
!****************************************
subroutine AddImageArray(set, name, image, imageArray, remove)
    type (DataSetT), intent(in) :: set
 
    character(*), intent(in) :: name
    type(imageT), intent(in) :: image
    real(single), intent(in), dimension &
               (0 : image%dims(0) - 1, 0 : image%dims(1) - 1) :: imageArray
    logical, intent(in) :: remove
    type (ArrayT) :: arr
    real (single), dimension(:, :), pointer :: optr
    integer(int32) :: i, j, k   


    arr = addArray(set, name, REAL32, dimensions = image%dims)
    optr=>real32Array2Data(arr)
 
    !******************************************
    ! Set pixels outside image boundary to NULL
    ! Have to set all the pixels to NULL to 
    ! begin with then set positive pixels
    !******************************************


    k = 0
    do i = 0, image%dims(0) - 1
        do j = 0, image%dims(1) - 1
            call setToNull(arr, k)
           
            k = k + 1
        end do
    end do 
 

  
    do i = 0, image%dims(0) - 1
        do j = 0, image%dims(1) - 1
            if (imageArray(i, j) .ge. 0.0) then
                optr(i, j) = imageArray(i, j)
            end if
            
        end do
    end do 
 
   
    !***********************************************
    ! Remove gaps in image
    !**********************************************
    if(remove) call RemoveGaps(image, imageArray, optr)

    ! Add WCS keywords
  
    call setAttribute(arr, 'CTYPE1', 'RA---TAN',&
         & 'Coord. type: RA tangent plane projection')
    call setAttribute(arr, 'CRPIX1', image%xrpix, 'WCS reference pixel')
    call setAttribute(arr, 'CRVAL1', image%xrval,&
         & 'coord. at X axis ref. pixel')
    call setAttribute(arr, 'CDELT1', image%xdelt, 'WCS pixel size')
    call setAttribute(arr, 'CTYPE2', 'DEC--TAN',&
         & 'Coord. type: DEC tangent plane projection')
    call setAttribute(arr, 'CRPIX2', image%yrpix, 'WCS reference pixel')
    call setAttribute(arr, 'CRVAL2', image%yrval,&
         & 'coord. at Y axis ref. pixel')
    call setAttribute(arr, 'CDELT2', image%ydelt, 'WCS pixel size')
    call setAttribute(arr, 'POSCOROK', image%poscorok, 'Astrometry correction applied')
    call setAttribute(arr, 'POSINTRP', image%posintrp, 'Astrom. corr. by interpolation')    
    ! Release dataset
    call release(arr)

end subroutine AddImageArray

!*******************************************************
! CompareDates - Compares the first date with the second.
! Returns -1 if the second is less than the first,
! 0 if they are equal and 1 if greater.
!*******************************************************
integer function CompareDates(dateString1, dateString2)
    character(*), intent(in) :: dateString1, dateString2
    integer :: time1(6), time2(6)        ! Stores the time from the date string
    character(1) char    ! Used for date and time extraction 
    integer i

    read(dateString1, '(i4, a1, i2, a1, i2, a1, i2, a1, i2, a1, i2)') &
         time1(1), char, time1(2), char, time1(3), char, time1(4), &
         char, time1(5), char, time1(6)

    read(dateString2, '(i4, a1, i2, a1, i2, a1, i2, a1, i2, a1, i2)') &
         time2(1), char, time2(2), char, time2(3), char, time2(4), &
         char, time2(5), char, time2(6)

    do i = 1, 6

        CompareDates = time2(i) - time1(i)
        if (CompareDates .ne. 0) return
 
    end do

    CompareDates = 0 ! Must be equal

end function CompareDates

!*************************************************************
! Subroutine GetObservationDetails
! Retrieves the observation details from a fits file
! for plottiong on the graph
!************************************************************
subroutine GetObservationDetails(dataSet, obsDetails)
    type(DataSetT), intent(in) :: dataSet
    type(OBSERVATIONDETAILS), intent(inout) :: obsDetails  
 

    call GetStringAttribute(dataSet, "TELESCOP", obsDetails%telescope)
    call GetStringAttribute(dataSet, "INSTRUME", obsDetails%instrument)
    call GetStringAttribute(dataSet, "OBSERVER", obsDetails%observer)
    call GetStringAttribute(dataSet, "OBJECT", obsDetails%object)


    call GetStringAttribute(dataSet, "FILTER", obsDetails%filter)
 
    call GetStringAttribute(dataSet, "OBS_ID",   obsDetails%obsID)
    call GetStringAttribute(dataSet, "EXP_ID",   obsDetails%expID)
    call GetStringAttribute(dataSet, "DATE-OBS", obsDetails%startObs)
    call GetStringAttribute(dataSet, "DATE-END", obsDetails%endObs)
    call GetRealAttribute  (dataSet, "EXPOSURE", obsDetails%exposure)


end subroutine GetObservationDetails

!***********************************************************
! Subroutine GetStringAttribute - 
!            Attempts to read a string attribute
!                        
!***********************************************************
!subroutine GetStringAttribute(dataSet, attribute, string)
!    type(DataSetT), intent(in) :: dataSet
!    character(*), intent(in) :: attribute
!    character(*), intent(inout) :: string!
!
!    if (.not. hasAttribute(dataSet, attribute)) then
!        string = ""
!    else
!        string = trim(stringAttribute(dataSet, attribute))
!    end if
!end subroutine GetStringAttribute

!***********************************************************
! Subroutine GetRealAttribute - 
!         Attempts to read a real attribute
!                        from a fits file
!                      
!***********************************************************
subroutine GetRealAttribute(dataSet, attribute, number)
    type(DataSetT), intent(in) :: dataSet
    character(*), intent(in) :: attribute
    real, intent(inout) :: number

    if (.not. hasAttribute(dataSet, attribute)) then
        number= 0.0
    else
        number = real32Attribute(dataSet, attribute)
    end if
   
end subroutine GetRealAttribute

!********************************************************
! Subroutine AddFitsKeyWords
!*******************************************************
subroutine AddFitsKeyWords(set, obsDetails)
    type (DataSetT), intent(in) :: set
    type (OBSERVATIONDETAILS), intent(in) :: obsDetails

    !****************************************
    ! Add various keywords to set.
    !****************************************
    call setAttribute(set, "RADECSYS", "FK5", " ")
    call setAttribute(set, "EQUINOX", 2000, "", "Equinox of celestial coordinate system")
    call setAttribute(set, "EXPOSURE", obsDetails%exposure, "secs", "Exposure")
    call setAttribute(set, "DATE-OBS", obsDetails%startObs, "Start date of the first exposure")
    call setAttribute(set, "DATE-END", obsDetails%EndObs,   "End date of the last exposure")
    call setAttribute(set, "OBSERVER", obsDetails%observer, "Name of PI")
    call setAttribute(set, "OBJECT",   obsDetails%object, "Name of object")


    call setAttribute(set, "INSTRUME", obsDetails%instrument, "Instrument")
    call setAttribute(set, "FILTER", obsDetails%filter, "Filter")
!    call setAttribute(set, "POSCOROK", obsDetails%poscorok, "Astrometry correction applied")    
 
!    call OAL_addCommonAttributes(set)


end subroutine AddFitsKeyWords





!***************************************************
! Subroutine RemoveGaps
!***************************************************
subroutine RemoveGaps(image, imageArray, optr)
    type(imageT), intent(in) :: image
    real(single), intent(in), dimension &
               (0 : image%dims(0) - 1, 0 : image%dims(1) - 1) :: imageArray
   
    real (single), dimension(:, :), pointer :: optr
    integer(int32) :: i, j, k, l, m  
      
    integer :: jStart = -1, jFinish = -1, nSize
    logical :: start
    integer, parameter :: maxSize = 100, width = 2
    real(single), pointer :: data(:) => null()
 


    allocate(data(1 : maxSize))

    do i = 3, image%dims(0) - 4
        jStart = -1
        start = .false.
        nSize = 0

        do j = 3, image%dims(1) - 4
            
            if (imageArray(i, j) .ge. 0.0) then
                start = .true.

                optr(i, j) = imageArray(i, j)
                if (jStart .gt. 0 .and. jFinish .ge. jStart) then ! Replace nan pixel value with median of surrounding pixels
                    if (jFinish - jStart .lt. 12) then
                        do k = jStart, jFinish
                            nSize = 0
                            lLoop : do l = max(0, k - width), min(k + width, image%dims(1) - 1)
                          
                                do m = max(0, i - width), min(i + width, image%dims(0) - 1)
                                    if(imageArray(m, l) .ge. 0.0) then
                                        nSize = nSize + 1
                                        data(nSize) = imageArray(m, l)
                                        if (nSize .eq. maxSize) exit lLoop
                                    end if
                                end do
                            end do lLoop
                            if (nSize .gt. 2) then  
                                optr(i, k) = MedianValue1(data, nSize)
                                 
                            else
                                optr(i, k) = (imageArray(i, jStart - 1) + imageArray(i, jFinish + 1)) / 2.0
                                
                            end if  
                        end do

                        jStart = -1
                        jFinish = -1    
                    else
                        jStart = -1
                        jFinish = -1
                       
                    end if
                end if 
            else
                if (jStart .lt. 0 .and. start) then
                    jStart = j
                    jFinish = j
                else
                    jFinish = j
                end if  

            end if
       
        end do


    end do 
    deallocate(data)
end subroutine RemoveGaps   

!********************************************************************
! subroutine StringDateToYear
! converts string values of the start and end time of the pbservation
! in the format yyyy-mm-ddThh:mm:ss.sss to the double precision 
! values as fraction of the year
!********************************************************************
real(double) function StringDateToYear(stringDateObs)
    character(40), intent(in) :: stringDateObs
    real(double) :: year
    integer :: tpos
    character(4) :: sYear
    character(2) :: sMonth, sDay
    character(2) :: sHour,sMinute
    character(6) :: sSecond
    year=0.0_double
    tpos=index(stringDateObs,"T")
    !write(*,*) "L1246 tpos=", tpos
    if (tpos .gt. 10) then
       sYear=stringDateObs(tpos-10:tpos-10+3)
       sMonth=stringDateObs(tpos-10+5:tpos-10+6)
       sDay=stringDateObs(tpos-10+8:tpos-10+9)
       sHour=stringDateObs(tpos+1:tpos+2)
       sMinute=stringDateObs(tpos+4:tpos+5)
       sSecond=stringDateObs(tpos+7:tpos+12)
    endif
    !write(*,*) sYear, "  ", sMonth, "  ", sDay, "     ", sHour, " ", &
    !     sMinute, " ", sSecond 
    StringDateToYear=year
end function StringDateToYear


end module ommosaic_mod









